(********************************************************)
(*   Olivier Boite (under supervision C. Dubois) - 2001 *)
(*   Catherine Dubois - translation to V8 - sept. 2012  *)
(********************************************************)

Require Import Language_spec.
Require Import Table.
Require Import Typing_spec.
Require Import Small_step_spec.
Require Import Substitution_lemma_proof.




Definition extend (env' env : env_typ) : Prop :=
  match env', env with
  | (envx', envl'), (envx, envl) =>
      envx' = envx /\
      (forall (l : location) (t : type),
       (in_dom l envl -> in_dom l envl') /\
       (apply_table envl l t -> apply_table envl' l t))
  end.


(*-----------------------------------*)
Lemma extend_id : forall env : env_typ, extend env env.
intro env.
elim env.
unfold extend in |- *.
intros envx envl.
split.
trivial.
intros.
split; (intro; assumption).
Qed.


(*-----------------------------------*)
Lemma extend_add :
 forall (envx : env_typ_var) (envl : env_typ_loc) (l : location) (t : type),
 ~ in_dom l envl -> extend (envx, add_table l t envl) (envx, envl).
intros envx envl.
elim envl.
intros dom f l t l_not_in_dom.
unfold extend in |- *.
split.
trivial.

intros l0 t0.
split.
intros l0_in_dom.
apply in_dom_add; assumption.

intro Happ.
case (eq_A_dec l l0).
intro Heq.
rewrite Heq.
inversion Happ.
simpl in l_not_in_dom.
rewrite Heq in l_not_in_dom.
absurd (in_prop l0 dom); assumption.

intro Hneq.
apply apply_add_non; assumption.
Qed.


 
(*-----------------------------------*)
Definition memory_respects_env (env : env_typ) (m : memory) : Prop :=
  match env, m with
  | (envx, envl), intro_table domm fm =>
      (forall (l : location) (t : type),
       in_prop l domm ->
       apply_table envl l t -> type_of env (value2expr (fm l)) t) /\
      match envl with
      | intro_table doml fl => domm = doml
      end
  end.




(*-----------------------------------*)
Lemma env_loc_useless :
 forall (e : expr) (l0 : location) (t t' : type) (envx : env_typ_var)
   (envl : env_typ_loc),
 ~ in_dom l0 envl ->
 type_of (envx, envl) e t' -> type_of (envx, add_table l0 t envl) e t'.

intros e.
induction e
 as
  [c|
   i|
   l|
   i e Hrece|
   e1 Hrece1 e0 Hrece0|
   i e1 Hrece1 e0 Hrece0|
   e1 Hrece1 e0 Hrece0 e2 Hrece2|
   e1 Hrece1 e0 Hrece0|
   e Hrece|
   e Hrece|
   e1 Hrece1 e0 Hrece0|
   e1 Hrece1 e0 Hrece0]; intros l0 t t' envx envl l_not_in_envl Htyp;
 inversion Htyp.

(*const*)
apply type_of_const.
(*var*)
apply type_of_var; assumption.
(*loc*)
apply type_of_loc.
case (eq_A_dec l l0).
(*l=l0*)
intro l_is_l0.
rewrite <- l_is_l0.
inversion H3.
rewrite <- H5 in l_not_in_envl; simpl in l_not_in_envl.
rewrite l_is_l0 in H4; absurd (in_prop l0 dom); assumption.

(*not l=l0*)
intro l_isnt_l0.
apply apply_add_non; auto.

(*fun*)
apply type_of_fun.
apply Hrece; assumption.

(*apply*)
apply type_of_app with (t := t0).
apply Hrece1; assumption.
apply Hrece0; assumption.

(*let*)
apply type_of_let with (t := t0).
apply Hrece1; assumption.
apply Hrece0; assumption.

(*if*)
apply type_of_if with (t := t').
apply Hrece1; assumption.
apply Hrece0; assumption.
apply Hrece2; assumption.

(*while*)
apply type_of_while with (t := t0).
apply Hrece1; assumption.
apply Hrece0; assumption.

(*ref*)
apply type_of_ref.
apply Hrece; assumption.

(*deref*)
apply type_of_deref.
apply Hrece; assumption.

(*sequ*)
apply type_of_sequ with (t' := t'0).
apply Hrece1; assumption.
apply Hrece0; assumption.

(*assign*)
apply type_of_assign with (t := t0).
apply Hrece1; assumption.
apply Hrece0; assumption.
Qed.


(***************************************)
Lemma add_respects_env :
 forall (envx : env_typ_var) (envl : env_typ_loc) (l : location) 
   (m : memory) (t : type) (v : expr) (V : is_value v),
 memory_respects_env (envx, envl) m ->
 type_of (envx, envl) v t ->
 ~ in_dom l m ->
 memory_respects_env (envx, add_table l t envl)
   (add_table l (value_intro v V) m).

intros envx envl l m t v V.
elim m.
intros domm fm.
elim envl.
intros doml fl.
intro m_respects_env; inversion_clear m_respects_env.
intros Htyp_v l_not_in_dom.

rewrite H0.
set (m' := add_table l (value_intro v V) (intro_table doml fm)) in |- *. 
simpl in (value of m').
unfold m' in |- *.
case (in_dec l doml).

simpl in l_not_in_dom; intros l_in_dom; rewrite H0 in l_not_in_dom;
 absurd (in_prop l doml); assumption.

intros l_not_in_dom_bis.

case (in_dec l doml).
simpl in l_not_in_dom; intros l_in_dom; rewrite H0 in l_not_in_dom;
 absurd (in_prop l doml); assumption.

intros l_not_in_dom_ter.

split.
intros l0 t0 l0_in_l_cons_domm Happ.
case (eq_A_dec l l0).

(*l=l0*)
intro l_is_l0.
cut (value2expr (value_intro v V) = v).
intro simpl.
rewrite simpl.
cut (t = t0).
intro t_is_t0.
rewrite <- t_is_t0.
apply env_loc_useless.
simpl in |- *.
simpl in l_not_in_dom.
assumption.

assumption.

rewrite l_is_l0 in Happ.
apply add_table_correct with (a := l0) (t := intro_table doml fl); assumption.


simpl in |- *; trivial.

(* not l=l0 *)
intro l_isnt_l0.
(*
Cut (type_of (envx,(intro_table doml fl)) (value2expr (fm l0)) t0).
Intro Htyp_fm_l0.
*)
apply env_loc_useless.
simpl in |- *.
(*Simpl in l_not_in_dom.*)
assumption.

apply H.
inversion l0_in_l_cons_domm.
absurd (l0 = l); auto.

rewrite H0; assumption.

(*Inversion Happ.*)
apply apply_add_non_bis with (a := l) (b := t).
assumption.

assumption.

(*Rewrite <- H1.
Apply intro_apply_table.
Assumption.

Apply H.
Inversion l0_in_l_cons_domm.
Absurd l0=l; Auto.

Rewrite H0;Assumption.

Inversion Happ.
Apply apply_add_non_bis with a:=l b:=t.
Assumption.

Rewrite <- H1.
Apply intro_apply_table.
Assumption.
*)

unfold add_table in |- *.
case (in_dec l doml).
intro; absurd (in_prop l doml); assumption.

intro; trivial.
Qed.




(*-----------------------------------*)
Lemma memory_respects_env_grows :
 forall (l : location) (v : expr) (t : type) (V : is_value v)
   (envx : env_typ_var) (envl : env_typ_loc) (m : memory),
 type_of (envx, envl) v t ->
 apply_table envl l t ->
 memory_respects_env (envx, envl) m ->
 memory_respects_env (envx, envl) (add_table l (value_intro v V) m).

intros l v t V envx envl m.
elim m.
intros domm fm.
elim envl.
intros doml fl Htyp Happ Hrespects.
simpl in |- *.
case (in_dec l domm).
(*l_in_domm*)
intro l_in_domm.
split.
intros l0 t0 l0_in_domm Happ_l0.
case (eq_A_dec l l0).

(* l=l0 *)
intro l_is_l0.
inversion Happ_l0.
rewrite l_is_l0 in Happ; inversion Happ.
rewrite H7; simpl in |- *; assumption.

(* not l=l0 *)
intro l_isnt_l0.
inversion Hrespects.
apply H; assumption.

inversion Hrespects; assumption.


(*not l_in_domm*)
intro l_not_in_domm.
inversion Hrespects.
rewrite <- H0 in Happ.
inversion Happ.
absurd (in_prop l domm); assumption.
Qed.



(* ------------------------*)
Definition type_of_config (env : env_typ) (s : configuration) 
  (t : type) :=
  match s with
  | (a, m) => type_of env a t /\ memory_respects_env env m
  end.


(*-----------------------------------*)
Definition less_typable (s1 s2 : configuration) : Prop :=
  forall (env : env_typ) (t : type),
  type_of_config env s1 t ->
  exists env' : env_typ, extend env' env /\ type_of_config env' s2 t.


(* ---------------------TYPE_PRESERVATION_BY_RED_EPSILON---------------------------- *)
Lemma type_preservation_by_red_epsilon :
 forall s1 s2 : configuration, red_epsilon s1 s2 -> less_typable s1 s2.
unfold less_typable in |- *.
intros.
inversion H.

(*(Fun x e) v -> e[x\v]*)
exists env.
split.
apply extend_id.

rewrite <- H3 in H0; inversion H0.
inversion H5.
unfold type_of_config in |- *.
split.

inversion H10.
rewrite <- H14 in H12.
apply substitution_lemma with (e1 := e) (i := x) (e := v) (t := t0);
 assumption.

assumption.

(*let_in -> e[x\v]*)
exists env.
split.
apply extend_id.

rewrite <- H3 in H0; inversion H0.
inversion H5.
unfold type_of_config in |- *.
split.

apply substitution_lemma with (e1 := e) (i := x) (e := v) (t := t0);
 assumption.

rewrite H9; assumption.

(*ref*)
rewrite <- H2 in H0; inversion H0.
inversion H4.
induction env as (a, b).
exists (a, add_table l t0 b).

split.
apply extend_add.
induction b as (l0, b).
induction m as (l1, b0).
inversion H5.
rewrite H11 in H1.
simpl in |- *.
simpl in H1.
assumption.

unfold type_of_config in |- *.
split.

apply type_of_loc.
apply apply_add.
apply add_respects_env.
assumption.

assumption.

simpl in |- *; assumption.

(*deref*)
exists env.
split.
apply extend_id.
 
rewrite <- H3 in H0; inversion_clear H0.
unfold type_of_config in |- *.
split.

inversion_clear H5.
inversion H0.
inversion H9.

rewrite <- H7 in H6; rewrite <- H11 in H6.
inversion H2.
rewrite <- H15 in H6.

unfold memory_respects_env in H6.
rewrite <- H11 in H9.
rewrite H13.
inversion H6.
apply H18; assumption.

assumption.

(*assign*)
exists env.
split.
apply extend_id.
rewrite <- H2 in H0; inversion_clear H0.
inversion H4.
inversion H8.
unfold type_of_config in |- *.
split.

replace Unit with (type_of_constant unit).
apply type_of_const.
simpl in |- *; trivial.

apply memory_respects_env_grows with (t := t0).
rewrite H12; assumption.

assumption.

rewrite H12; assumption.

(*sequ*)
exists env.
split.
apply extend_id.

rewrite <- H2 in H0; inversion H0.
inversion H4.
unfold type_of_config in |- *; split.
assumption.
assumption.

(*if_true*)
exists env.
split.
apply extend_id.

rewrite <- H1 in H0; inversion H0.
unfold type_of_config in |- *; split.
inversion H3.
assumption.
assumption.

(*if_false*)
exists env.
split.
apply extend_id.

rewrite <- H1 in H0; inversion H0.
unfold type_of_config in |- *; split.
inversion H3.
assumption.
assumption.


(* while*)
exists env.
split.
apply extend_id.

rewrite <- H1 in H0; inversion H0.
unfold type_of_config in |- *; split.
inversion H3.

apply type_of_if.
assumption.

apply type_of_sequ with (t' := t0).
assumption.

apply type_of_while with (t := t0).
assumption.

assumption.

replace Unit with (type_of_constant unit).
apply type_of_const.

simpl in |- *.
trivial.

assumption.

Qed.



(*-------------------------------------------------------------------------*)

Lemma type_of_extend :
 forall (e : expr) (env env' : env_typ) (t : type),
 type_of env e t -> extend env' env -> type_of env' e t.

intro e;
 induction e
  as
   [c|
    i|
    l|
    i e Hrece|
    e1 Hrece1 e0 Hrece0|
    i e1 Hrece1 e0 Hrece0|
    e1 Hrece1 e0 Hrece0 e2 Hrece2|
    e1 Hrece1 e0 Hrece0|
    e Hrece|
    e Hrece|
    e1 Hrece1 e0 Hrece0|
    e1 Hrece1 e0 Hrece0]; intros env env' t; elim env; 
 intros envx envl; elim env'; intros envx' envl'; elim envl';
 intros doml' fl' Htyp Hext; inversion Hext; rewrite H; 
 inversion Htyp.

(*const*)
apply type_of_const.

(*var*)
apply type_of_var; assumption.

(*loc*)
apply type_of_loc.
elim (H0 l t0); auto.

(*fun*)
apply type_of_fun.
apply Hrece with (env := (add_table i t0 envx, envl)).
assumption.

split.
trivial.

assumption.

(*apply*)
apply type_of_app with (t := t0).
apply Hrece1 with (env := (envx, envl)).
assumption.

split.
trivial.

assumption.

apply Hrece0 with (env := (envx, envl)).
assumption.

split.
trivial.

assumption.

(*let*)
apply type_of_let with (t := t0).
apply Hrece1 with (env := (envx, envl)).
assumption.

split.
trivial.

assumption.

apply Hrece0 with (env := (add_table i t0 envx, envl)).
assumption.

split.
trivial.

assumption.

(*if*)
apply type_of_if.
apply Hrece1 with (env := (envx, envl)).
assumption.

split.
trivial.

assumption.

apply Hrece0 with (env := (envx, envl)).
assumption.

split.
trivial.

assumption.

apply Hrece2 with (env := (envx, envl)).
assumption.

split.
trivial.

assumption.

(*while*)
apply type_of_while with (t := t0).
apply Hrece1 with (env := (envx, envl)).
assumption.

split.
trivial.

assumption.

apply Hrece0 with (env := (envx, envl)).
assumption.

split.
trivial.

assumption.

(*ref*)
apply type_of_ref.
apply Hrece with (env := (envx, envl)).
assumption.

split.
trivial.

assumption.

(*deref*)
apply type_of_deref.
apply Hrece with (env := (envx, envl)).
assumption.

split.
trivial.

assumption.

(*sequ*)
apply type_of_sequ with (t' := t').
apply Hrece1 with (env := (envx, envl)).
assumption.

split.
trivial.

assumption.

apply Hrece0 with (env := (envx, envl)).
assumption.

split.
trivial.

assumption.

(*assign*)
apply type_of_assign with (t := t0).
apply Hrece1 with (env := (envx, envl)).
assumption.

split.
trivial.

assumption.

apply Hrece0 with (env := (envx, envl)).
assumption.

split.
trivial.

assumption.
Qed.


(**************************************)
Lemma extend_var :
 forall (envx envx' : env_typ_var) (envl envl' : env_typ_loc)
   (x : identifier) (t : type),
 extend (envx', envl') (envx, envl) ->
 extend (add_table x t envx', envl') (add_table x t envx, envl).
intros envx envx' envl envl' x t Hext.
inversion Hext.
split.
rewrite H; trivial.

assumption.

Qed.


(*--------------------------------*)
Lemma less_typable_grows :
 forall (c : context) (s1 s2 : configuration),
 is_context c ->
 less_typable s1 s2 -> less_typable (app_ctx c s1) (app_ctx c s2).
intros c s1 s2 H.
elim s1.
elim s2.
intros a2 m2 a1 m1 H0.
induction H
 as
  [|
   e c H HrecH|
   v c H HrecH H1|
   e y c H HrecH|
   e c H HrecH|
   c H HrecH|
   c H HrecH|
   l c H HrecH|
   e c H HrecH|
   e1 e2 c H HrecH]; unfold app_ctx in |- *; try unfold app_ctx in HrecH.

(*hole*)
assumption.

(* (c a1) e *)
unfold less_typable in |- *; intros.
inversion H1.
inversion H2.
unfold less_typable in HrecH.
cut (type_of_config env (c a1, m1) (Arrow t0 t)).
intro.
elim (HrecH env (Arrow t0 t) H10).
intros env' Hind.
inversion_clear Hind.
unfold type_of_config in H12; inversion_clear H12.

exists env'.
split.
assumption.

unfold type_of_config in |- *; split.
apply type_of_app with (t := t0).
assumption.
apply type_of_extend with (env := env); assumption.
assumption.
unfold type_of_config in |- *; split.
assumption.
assumption.

(* v (c a1) *)
unfold less_typable in |- *; intros.
inversion H2.
inversion H3.
unfold less_typable in HrecH.
cut (type_of_config env (c a1, m1) t0).
intro.
elim (HrecH env t0 H11).
intros env' Hind.
inversion_clear Hind.
unfold type_of_config in H13; inversion_clear H13.
exists env'.
split.
assumption.


unfold type_of_config in |- *; split.

apply type_of_app with (t := t0).
apply type_of_extend with (env := env); assumption.
assumption.
assumption.

unfold type_of_config in |- *; split.
assumption.
assumption.

(*let*)
unfold less_typable in |- *; intros.
inversion H1.
inversion H2.
unfold less_typable in HrecH.
cut (type_of_config env (c a1, m1) t0).
intro.
elim (HrecH env t0 H11).
intros env' Hind.
inversion_clear Hind.
unfold type_of_config in H13; inversion_clear H13.
exists env'.
split.
rewrite H6; assumption.

unfold type_of_config in |- *; split.

rewrite <- H6 in H12.

induction env' as (a, b).
apply type_of_let with (t := t0).
assumption.
apply type_of_extend with (env := (add_table y t0 envx, envl)).
assumption.
apply extend_var.
assumption.

assumption.

unfold type_of_config in |- *; split.
rewrite <- H6; assumption.
assumption.


(*sequ (c e) a1 *)
unfold less_typable in |- *; intros.
inversion H1.
inversion H2.
unfold less_typable in HrecH.
cut (type_of_config env (c a1, m1) t').
intro.
elim (HrecH env t' H10).
intros env' Hind.
inversion_clear Hind.
unfold type_of_config in H12; inversion_clear H12.
exists env'.
split.

assumption.

unfold type_of_config in |- *; split.
apply type_of_sequ with (t' := t').
assumption.

apply type_of_extend with (env := env); assumption.

assumption.

unfold type_of_config in |- *; split.
assumption.
assumption.

(*ref*)
unfold less_typable in |- *; intros.
inversion H1.
inversion H2.
unfold less_typable in HrecH.
cut (type_of_config env (c a1, m1) t0).
intro.
elim (HrecH env t0 H8).
intros env' Hind.
inversion_clear Hind.
unfold type_of_config in H10; inversion_clear H10.
exists env'.
split.

assumption.

unfold type_of_config in |- *; split.
apply type_of_ref.
assumption.

assumption.

unfold type_of_config in |- *; split.
assumption.
assumption.

(*deref*)
unfold less_typable in |- *; intros.
inversion H1.
inversion H2.
unfold less_typable in HrecH.
cut (type_of_config env (c a1, m1) (Ref_t t)).
intro.
elim (HrecH env (Ref_t t) H8).
intros env' Hind.
inversion_clear Hind.
unfold type_of_config in H10; inversion_clear H10.
exists env'.
split.

assumption.

unfold type_of_config in |- *; split.
apply type_of_deref.
assumption.

assumption.

unfold type_of_config in |- *; split.
assumption.
assumption.

(*assign right*)
unfold less_typable in |- *; intros.
inversion H1.
inversion H2.
unfold less_typable in HrecH.
cut (type_of_config env (c a1, m1) t0).
intro.
elim (HrecH env t0 H10).
intros env' Hind.
inversion_clear Hind.

unfold type_of_config in H12; inversion_clear H12.
exists env'.
split.

assumption.

unfold type_of_config in |- *; split.
apply type_of_assign with (t := t0).
apply type_of_extend with (env := env); assumption.
assumption.

assumption.

unfold type_of_config in |- *; split.
assumption.
assumption.

(*assign left*)
unfold less_typable in |- *; intros.
inversion H1.
inversion H2.
unfold less_typable in HrecH.
cut (type_of_config env (c a1, m1) (Ref_t t0)).
intro.
elim (HrecH env (Ref_t t0) H10).
intros env' Hind.
inversion_clear Hind.

unfold type_of_config in H12; inversion_clear H12.
exists env'.
split.

assumption.

unfold type_of_config in |- *; split.
apply type_of_assign with (t := t0).
assumption.
apply type_of_extend with (env := env); assumption.

assumption.

unfold type_of_config in |- *; split.
assumption.
assumption.

(*if*)
unfold less_typable in |- *; intros.
inversion H1.
inversion H2.
unfold less_typable in HrecH.
cut (type_of_config env (c a1, m1) (Const_t Bool)).
intro.
elim (HrecH env (Const_t Bool) H12).
intros env' Hind.
inversion_clear Hind.

unfold type_of_config in H14; inversion_clear H14.
exists env'.
split.

assumption.

unfold type_of_config in |- *; split.
apply type_of_if.
assumption.
apply type_of_extend with (env := env); assumption.
apply type_of_extend with (env := env); assumption.

assumption.

unfold type_of_config in |- *; split.
assumption.
assumption.

Qed.


(*-------------   TYPE IS PRESERVED  BY REDUCTION  -----------*)
Theorem Subject_reduction :
 forall s1 s2 : configuration, red s1 s2 -> less_typable s1 s2.
simple induction 1; intros.
apply type_preservation_by_red_epsilon; assumption.
 
apply less_typable_grows; assumption.   
Qed.                                                                             
