Require Import Language_spec. 
Require Import Table.

Parameter stamp : Set.

Inductive Basic : Set :=
  | Int : Basic
  | Bool : Basic
  | Unit : Basic.
 
Inductive type : Set :=
  | Const_t : Basic -> type
  | Var_t : stamp -> type
  | Arrow : type -> type -> type
  | Ref_t : type -> type.


Definition type_of_constant (c : Constante) : Basic :=
  match c with
  | Const_int i => Int
  | Const_bool b => Bool
  | unit => Unit
  end.



Definition env_typ_var := table identifier type.
Definition env_typ_loc := table location type.
Definition env_typ := (env_typ_var * env_typ_loc)%type.


Inductive type_of : env_typ -> expr -> type -> Prop :=
  | type_of_const :
      forall (env : env_typ) (c : Constante),
      type_of env (Const c) (Const_t (type_of_constant c))
  | type_of_var :
      forall (envx : env_typ_var) (envl : env_typ_loc) 
        (x : identifier) (t : type),
      apply_table envx x t -> type_of (envx, envl) (Var x) t
  | type_of_app :
      forall (env : env_typ) (e e' : expr) (t t' : type),
      type_of env e (Arrow t t') ->
      type_of env e' t -> type_of env (Apply e e') t'
  | type_of_fun :
      forall (envx : env_typ_var) (envl : env_typ_loc) 
        (x : identifier) (e : expr) (t t' : type),
      type_of (add_table x t envx, envl) e t' ->
      type_of (envx, envl) (Fun x e) (Arrow t t')
  | type_of_let :
      forall (x : identifier) (envx : env_typ_var) 
        (envl : env_typ_loc) (e e' : expr) (t t' : type),
      type_of (envx, envl) e t ->
      type_of (add_table x t envx, envl) e' t' ->
      type_of (envx, envl) (Let_in x e e') t'
  | type_of_if :
      forall (env : env_typ) (e e1 e2 : expr) (t : type),
      type_of env e (Const_t Bool) ->
      type_of env e1 t -> type_of env e2 t -> type_of env (If e e1 e2) t
  | type_of_while :
      forall (env : env_typ) (e1 e2 : expr) (t : type),
      type_of env e1 (Const_t Bool) ->
      type_of env e2 t -> type_of env (While e1 e2) (Const_t Unit)
  | type_of_ref :
      forall (env : env_typ) (e : expr) (t : type),
      type_of env e t -> type_of env (Ref e) (Ref_t t)
  | type_of_deref :
      forall (env : env_typ) (e : expr) (t : type),
      type_of env e (Ref_t t) -> type_of env (Deref e) t
  | type_of_sequ :
      forall (env : env_typ) (e1 e2 : expr) (t t' : type),
      type_of env e1 t' -> type_of env e2 t -> type_of env (Sequ e1 e2) t
  | type_of_assign :
      forall (env : env_typ) (e1 e2 : expr) (t : type),
      type_of env e1 (Ref_t t) ->
      type_of env e2 t -> type_of env (Assign e1 e2) (Const_t Unit)
  | type_of_loc :
      forall (envx : env_typ_var) (envl : env_typ_loc) 
        (l : location) (t : type),
      apply_table envl l t -> type_of (envx, envl) (Loc l) (Ref_t t).
